from standard_j.utils import *
import typing
import datetime
import os


class LogStatusType(Enum):
    INFO: EnumItem
    SUCCESS: EnumItem
    WARNING: EnumItem
    ERROR: EnumItem


# <editor-fold desc="Private Variables">
__output_files: list[str] = []
__min_quit_status: LogStatusType = LogStatusType.ERROR
__time_format: str = "%m/%d/%Y %H:%M:%S:%f"
__raw_format: str = "<status WARNING\t| mm/dd/YYYY HH:MM:SS:ffffff>\t"
__output_format: str = "<status {status}\t| {time}>\t"
__current_buffer: list[str] = []
# </editor-fold>


def get_min_quit_status():
    return __min_quit_status


def set_min_quit_status(new_status: LogStatusType):
    if type(new_status) != LogStatusType:
        raise StandardJError(f"Min quit status for the log must be of type <enum LogStatusType>; found: {new_status}")


def create_dir(path: str):
    """
    Recursively creates directories until the specified path can be created
    :param path: str Directory to create
    :return: None
    """
    parent = os.path.dirname(path)
    if parent != "" and not os.path.exists(parent):
        create_dir(parent)
    os.mkdir(path)


def create_log_file(folder: str, file: str, overwrite=False, append_output_file=True):
    """
    Method used to create a log file
    :param folder: str folder path
    :param file: str file name
    :param overwrite: Optional[bool] Only applicable if specified file already exists:
                      If True, overwrite the file. If False, raise error
    :param append_output_file: Optional[bool] If True, the file is appended to the output_files list
    :return: None
    :raises FileExistsError: If the specified file exists and overwrite is False
    """
    if not os.path.exists(folder):
        # If the folder does not yet exist, create it and any parent directories required to create it
        create_dir(folder)
    full_path = os.path.join(folder, file)
    if not overwrite:
        # If overwrite is False, attempt to create the file, if it already exists, a FileExistsError is raised
        open(full_path, "x").close()
    else:
        # If overwrite is True, delete the old file if it exists, then create the new one
        if os.path.exists(full_path):
            os.remove(full_path)
        open(full_path, "x").close()
    # If append_output_file is True, add the newly created file to the output_files list
    if append_output_file:
        __output_files.append(full_path)


def get_output_files() -> tuple[str, ...]:
    """
    A getter function for the current list of output files
    :return: tuple[str, ...] A tuple of output files that will be written to
    """
    return tuple(__output_files)


def reset_output_files() -> None:
    """
    A function that resets the current output files so that they can be reassigned
    """
    __output_files.clear()


def enable_console_output() -> None:
    """
    A function that enables console output when other files have been added to the output_files
    """
    if None not in __output_files:
        __output_files.append(None)


def disable_console_output() -> None:
    """
    A function that disables console output when other files have been added to the output_files
    NOTE: if no other output files are specified, this is ignored as the console is the defoult output stream
    """
    if None in __output_files:
        __output_files.remove(None)


def put(*values: object, sep: str = " ", status: LogStatusType = LogStatusType.INFO,
        files: typing.Iterable[str | None] = None) -> None:
    """
    A function used to put values to the current log files, or a specified list of files
    :param values: *object Objects to put to the log
    :param sep: Optional[str] Seperator used between each value when putting to the files
    :param status: Optional[LogStatusType] Status of the current log (See LogStatusType for details)
    :param files: Optional[str] A list of files to output to instead of the current output files. A value of None
                  Results in using the current output files
    :return: None
    :raises StandardJError: If the min_quit_status is reached
    """
    current_time = datetime.datetime.now().strftime(__time_format)
    for f in files if files is not None else __output_files if len(__output_files) > 0 else (None,):
        if isinstance(f, str):
            f = open(f, "a")
        print(__output_format.format(status=status.name, time=current_time), f"{sep.join(str(val) for val in values)}",
              sep="", file=f, flush=True)
        if f is not None:
            f.close()
    if __min_quit_status is not None and int(__min_quit_status) <= int(status):
        flush()
        raise StandardJError("<standard_j.log> min quit status reached.  See output file(s) for details")


def buffer(*values, sep: str = " ") -> None:
    """
    A function used to buffer outputs to the log.  These outputs will be put when the flush function is called, or
    when the min quit status is reached
    :param values: *object Objects to buffer for the log
    :param sep: Optional[str] Seperator used between each value when buffering for the log
    :return: None
    """
    __current_buffer.append(sep.join(str(val) for val in values))


def clear() -> None:
    """
    A function used to clear the current buffer
    :return: None
    """
    __current_buffer.clear()


def flush(end: str = "", sep: str = f"\n{''.join(' ' for _ in range(len(__raw_format)))}",
          status: LogStatusType = LogStatusType.INFO, files: typing.Iterable[str | None] = None) -> None:
    """
    A function used to flush the current buffer to the output files, or specified "files"
    :param end: Optional[str] String to be appended onto the last msg in the buffer
    :param sep: Optional[str] String to be used between each msg in the buffer
    :param status: Optional[LogStatusType] Status of the current log (See LogStatusType for details)
    :param files: Optional[list[str, None]] Files to output to instead of the current output files
    :return: None
    """
    if len(__current_buffer) == 0:
        return
    __current_buffer[-1] += end
    buf = tuple(__current_buffer)
    clear()
    put(*buf, sep=sep, status=status, files=files)



from standard_j.utils import *
import standard_j.log
from inspect import Signature, Parameter
import functools
import itertools
import types
import typing
import warnings


class ExperimentalWarning(UserWarning):
    pass


def experimental(func):
    """
    Function decorator used to denote functions that are experimental and therefore unstable
    :param func: FunctionType Function to decorate
    :return: FunctionType Unaltered function
    """
    warnings.simplefilter("always", ExperimentalWarning)
    warnings.warn(f"Import of experimental function: {func.__name__}", ExperimentalWarning)

    return func


def experimental_class(cls):
    """
    Class decorator used to denote classes that are experimental and therefore unstable
    :param cls: ClassType Class to decorate
    :return: ClassType Unaltered class
    """
    warnings.simplefilter("always", ExperimentalWarning)
    warnings.warn(f"Import of experimental class: {cls.__name__}", ExperimentalWarning)

    return cls


# <editor-fold desc="DEBUG">
DEBUG = False


if DEBUG:
    __OVERLOAD_DEBUG_FOLDER = "logs\\debug"
    __OVERLOAD_DEBUG_FILE = f"{__OVERLOAD_DEBUG_FOLDER}\\overload_log.txt"
    standard_j.log.create_log_file(__OVERLOAD_DEBUG_FOLDER, "overload_log.txt",
                                   overwrite=True, append_output_file=False)
# </editor-fold>


class DecipheredFunction(object):
    """
    A simple class used to get details about the signature of a function
    """
    def __init__(self, func: types.FunctionType):
        self.func = func
        self.inspect_sig = Signature.from_callable(func)

        self.name = func.__qualname__
        self.parameters = tuple(parameter for parameter in self.inspect_sig.parameters.values())

        self.p_only_args = []
        self.either_p_args = []
        self.either_kw_args = []
        self.kw_only_args = []
        self.var_p_args = None
        self.var_kw_args = None

        for parameter in self.parameters:
            if parameter.kind is Parameter.POSITIONAL_ONLY:
                self.p_only_args.append(parameter.name)
            elif parameter.kind is Parameter.POSITIONAL_OR_KEYWORD:
                if parameter.default is Parameter.empty:
                    self.either_p_args.append(parameter.name)
                else:
                    self.either_kw_args.append(parameter.name)
            elif parameter.kind is Parameter.KEYWORD_ONLY:
                self.kw_only_args.append(parameter.name)
            elif parameter.kind is Parameter.VAR_POSITIONAL:
                self.var_p_args = parameter
            elif parameter.kind is Parameter.VAR_KEYWORD:
                self.var_kw_args = parameter
            else:
                raise StandardJError(f"What did you even do? {parameter.name} : {parameter.kind}")

        self._base = (self.name, self.parameters)

    def __hash__(self):
        return self._base.__hash__()

    def __eq__(self, other):
        return self._base == other

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __str__(self):
        return f"DecipheredFunction(name: {self.name}, p_only: {self.p_only_args}, either_p: {self.either_p_args}, " \
               f"either_kw: {self.either_kw_args}, kw_only: {self.kw_only_args}, " \
               f"var_p: {self.var_p_args}, var_kw: {self.var_kw_args})"

    def __repr__(self):
        return str(self)

    def check(self, num_args, kwargs):
        if not len(self.p_only_args) <= num_args:
            if DEBUG:
                standard_j.log.buffer("Fail{1}", f"(p_only: {self.p_only_args})")
            return False
        if not (num_args - len(self.p_only_args) <= len(self.either_p_args) + len(self.either_kw_args) or
                self.var_p_args is not None):
            if DEBUG:
                standard_j.log.buffer("Fail{2}", f"(num_all_p: {len(self.p_only_args) + len(self.either_p_args)},",
                                      f"var_p: {self.var_p_args})")
            return False
        if not (set(filter(lambda key: key not in self.kw_only_args, kwargs)
                    ).issubset(itertools.chain(self.either_p_args, self.either_kw_args)) or
                self.var_kw_args is not None):
            if DEBUG:
                standard_j.log.buffer("Fail{3}", f"(kw_only: {self.kw_only_args},", f"either_p: {self.either_p_args},",
                                      f"either_kw: {self.either_kw_args},", f"var_kw: {self.var_kw_args})")
            return False
        if not set(self.either_p_args[num_args - len(self.p_only_args):]).issubset(kwargs):
            if DEBUG:
                standard_j.log.buffer("Fail{4}", f"(either_p: {self.either_p_args},", f"p_only: {self.p_only_args})")
            return False
        if DEBUG:
            standard_j.log.buffer("Succeed:", self)
        return True
        # DEBUG
        # p_only <= num_args
        # num_args - p_only <= all_either or var_p is not None
        # kwargs - kw_only issubset of all either OR var_kw is not None
        # all_p[num_args:] issubset of kwargs


__overloaded_functions: dict[str, list[DecipheredFunction]] = {}


@experimental
def overload(func: types.FunctionType):
    """
    A function decorator used to overload functions. NOTE: this decorator must be applied to all versions of this
    function, including the first and last instances
    :param func: The function that is to be overloaded
    :return: A wrapper function that will select the appropriate overloaded function based on the passed args and kwargs
    """
    func = DecipheredFunction(func)
    if func.name in __overloaded_functions.keys():
        __overloaded_functions[func.name].append(func)
    else:
        __overloaded_functions[func.name] = [func]

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        matches = []
        if DEBUG:
            standard_j.log.buffer(f"{func.name}(num_args: {len(args)}, kwargs: {kwargs.keys()})")
        for sig in __overloaded_functions[func.name]:
            if sig.check(len(args), kwargs.keys()):
                matches.append(sig)
        if DEBUG:
            standard_j.log.flush(files=(__OVERLOAD_DEBUG_FILE, None))

        if len(matches) == 0:
            if DEBUG:
                standard_j.log.buffer(f"{func.name}(num_args: {len(args)},",
                                      f"kwargs: [{', '.join(k for k in kwargs.keys())}])")
                standard_j.log.buffer("No matches")
                standard_j.log.flush(status=standard_j.log.LogStatusType.ERROR, files=(__OVERLOAD_DEBUG_FILE, None))
            raise StandardJError("Error in overload: no matching functions for call: ",
                                 f"(# args: {len(args)}, kwargs: [{', '.join(k for k in kwargs.keys())}])")
        if len(matches) == 1:
            if DEBUG:
                standard_j.log.buffer(f"{func.name}(num_args: {len(args)}, kwargs: {kwargs.keys()})")
                standard_j.log.buffer(f"Match: {matches[0]}")
                standard_j.log.flush(files=(__OVERLOAD_DEBUG_FILE, None))
            return matches[0](*args, **kwargs)
        if DEBUG:
            standard_j.log.buffer(f"{func.name}(num_args: {len(args)},", f"kwargs: {kwargs.keys()})")
            standard_j.log.buffer(f"matches[{len(matches)}]:", matches)
            standard_j.log.flush(status=standard_j.log.LogStatusType.ERROR, files=(__OVERLOAD_DEBUG_FILE, None))
        raise StandardJError(f"Error in overload: too many[{len(matches)}] matching functions for call: ",
                             f"(# args: {len(args)}, kwargs: [{', '.join(k for k in kwargs.keys())}])")

    return func_wrapper


__recursion_guarded_functions = {}


@experimental
def recursion_guarded(func: typing.Callable, max_recursions: int = 0):
    """
    Function decorator that ensures the passed function will not be called recursively. Any would be recursive calls
    are ignored.  If Optional[max_limit] is provided, that is the maximum number of recursive calls allowed.

    NOTE: This function only checks for simultaneous calls to a function, if it is called multiple times successively,
    then those calls will not be ignored.
    :param func: Callable Function to decorate
    :param max_recursions: int The maximum number of recursive calls that are allowed before they are ignored
    :return: Callable Wrapper function
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if func not in __recursion_guarded_functions:
            __recursion_guarded_functions[func] = 0
        elif __recursion_guarded_functions[func] < max_recursions:
            __recursion_guarded_functions[func] += 1
        else:
            return
        output = func(*args, **kwargs)
        __recursion_guarded_functions[func] -= 1
        if __recursion_guarded_functions[func] == -1:
            __recursion_guarded_functions.pop(func)
        return output
    return wrapper


class SwitchCase(object):
    """
    Wrapper class for a lambda expression used in a switch statement. Used in case obj is a function type
    """
    def __init__(self, func: types.FunctionType | typing.Callable):
        self.func = func
        self.num_args = len(DecipheredFunction(func).parameters)

    def __call__(self, *args, **kwargs):
        if 0 < self.num_args:
            self.func(*args, **kwargs)
        else:
            self.func()


def switch(obj, cases: dict, error: Exception = None):
    """
    A functional form of the switch statement found in other languages.

    obj is checked against cases.keys() and if a match is found, the associated value is returned.
    A SwitchCase object can be used with a Function (or lambda) that returns a boolean as the parameter.

    If no match is found, an optional "DEFAULT" key is used.

    If no match is found and no "DEFAULT" key exists, a specified error will be raised or the function will return
    None if error is None
    :param obj: object to be checked
    :param cases: dict[cases_to_check, value_to_return] where each case can be an exact value, SwitchCase, or "DEFAULT"
    :param error: Exception to raise if no match is found
    :return: value_to_return from matching case or None
    :raises Exception: error is raised if no match is found and no "DEFAULT" key is specified
    """
    for case, value in cases.items():
        if case == obj:
            return value
        if isinstance(case, SwitchCase):
            if case(obj):
                return value
    if "DEFAULT" in cases.keys():
        return cases["DEFAULT"]
    if error is not None:
        raise error


def sc_switch(obj, cases: dict, error: Exception = None):
    """
    A modified version of the switch statement that allows for short-circuit evaluation.

    The values of the cases dict are placed within lambda functions to ensure they do not run
    unless that case matches obj.

    If no error is specified, the function exits without raising anything and None is returned
    :param obj: object to compare
    :param cases: dict[cases_to_match, lambda: val_to_return]
    :param error: Exception to raise if no match is found, and "DEFAULT" is not specified
    :return: val_to_return from lambda function for matching case or None if no match is found
    :raises Exception: error is raised if no match is found, and no "DEFAULT" is specified
    """
    out = switch(obj, cases, error)
    if out is not None:
        return out()
    return None



from setuptools import setup, find_packages
from standard_j import __version__

setup(
    name="standard_j",
    version=__version__,
    author="J7W539",
    author_email="j7w539@gmail.com",
    url="https://github.com/J7W539/standard-j",
    packages=find_packages(),
)



from standard_j.error import StandardJError
from typing import Self as EnumItem


class Initializer(object):
    """
    A class that tracks whether a subclass has been initialized (using inherited initialize())
    Useful for when methods need to act differently post initialization
    Generally best to initialize() at the end of __init__()
    """

    def __new__(cls, *args, **kwargs):
        """
        Used to ensure this class is only inherited, and not directly instantiated
        """
        if cls == Initializer:
            raise NotImplementedError("Initializer class is not implemented and must be inherited")
        return super().__new__(cls, *args, **kwargs)

    def initialize(self):
        """
        The method that initializes the object.
        :return: None
        """
        object.__setattr__(self, f"_{type(self).__name__}__initialized", 1)

    @property
    def is_initialized(self) -> bool:
        """
        Method used to find out if self has been initialized yet
        :return: True if self.__initialized exists, False otherwise
        """
        return f"_{type(self).__name__}__initialized" in object.__getattribute__(self, "__dict__").keys()


class __EnumMeta(type):
    class __EnumDict(dict):
        def __setitem__(self, key, value):
            if hasattr(self, "standard_keys") and key not in getattr(self, "standard_keys"):
                raise StandardJError("Enum attributes must be hinted, not set")
            super().__setitem__(key, value)

        def lock(self):
            if not hasattr(self, "standard_keys"):
                setattr(type(self), "standard_keys", self.keys())

    @classmethod
    def __prepare__(mcs, name, bases):
        return mcs.__EnumDict()

    def __new__(mcs, name, bases, e_dict: __EnumDict):
        enum_cls = super().__new__(mcs, name, bases, e_dict)
        e_dict.lock()

        if name != "Enum":
            if not all(value == EnumItem for value in e_dict["__annotations__"].values()):
                raise TypeError("All attributes of an Enum must be hinted as type 'EnumItem' or 'Self' (from typing)")
            annotations_list = list(e_dict["__annotations__"].keys())
            for index, enum_item in enumerate(annotations_list):
                type.__setattr__(enum_cls, enum_item, enum_cls(enum_item, index))
            type.__setattr__(enum_cls, f"_{enum_cls.__name__}__initialized", 1)
        return enum_cls

    def __iter__(self):
        for key in self.__annotations__.keys():
            yield super().__getattribute__(key)

    def __len__(self):
        return len(self.__annotations__.keys())

    def __setattr__(self, key, value):
        raise AttributeError("Cannot set the value of an Enum attribute")

    def __str__(self):
        return f"{self.__name__}({', '.join(item for item in self.__annotations__.keys())})"

    def __repr__(self):
        out = (f"{enum_item}: {index}" for index, enum_item in enumerate(self.__annotations__.keys()))
        return f"<enum {self.__name__}>({', '.join(out)})"


class Enum(metaclass=__EnumMeta):
    """
    Simple enum class for python that allows for typechecking members as self
    """
    __annotations__ = {}

    def __init__(self, name: str, value: int):
        if hasattr(self, f"_{type(self).__name__}__initialized"):
            raise TypeError(f"{str(self)} cannot be instantiated")
        self.name = name
        self.value = value
        object.__setattr__(self, f"_{type(self).__name__}__initialized", 1)

    def __getattribute__(self, item):
        try:
            super().__getattribute__(f"_{type(self).__name__}__initialized")
        except AttributeError:
            return super().__getattribute__(item)
        if item not in ("name", "value"):
            raise AttributeError("Enum attribute only has access to its 'name' and 'value'")
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if hasattr(self, f"_{type(self).__name__}__initialized"):
            raise AttributeError("Enum attribute cannot be modified")
        super().__setattr__(key, value)

    def __int__(self):
        return self.value

    def __str__(self):
        return f"{type(self).__name__}.{self.name}"

    def __repr__(self):
        return f"<{type(self).__name__} object>(name: {self.name}, value: {self.value})"
